// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/24DSI/24DSI_Linux_3.x_and_4.x_GSC_DN/utils/util_fsamp_ai.c $
// $Rev: 52688 $
// $Date: 2023-03-24 15:05:09 -0500 (Fri, 24 Mar 2023) $

// 24DSI: Utilities: source file

#include "main.h"



// macros *********************************************************************

#define	ABS(v)					(((v) < 0) ? (-(v)) : (v))
#define	LIMIT_MAX(v,max)		(((v) > (max)) ? (max) : (v))
#define	LIMIT_MIN(v,min)		(((v) < (min)) ? (min) : (v))
#define	LIMIT_RANGE(v,min,max)	(LIMIT_MIN((LIMIT_MAX((v),(max))),(min)))



// data types *****************************************************************

typedef struct
{
	s32		fsamp_want;
	double	fsamp_got;

	double	delta;
	double	fgen;
	s32		fgen_max;
	s32		fgen_min;
	s32		fref;
	s32		fsamp_max;
	s32		fsamp_min;
	s32		ndiv;
	s32		ndiv_max;
	s32		ndiv_min;
	s32		nrate;
	s32		nrate_max;
	s32		nrate_min;
	s32		nref;
	s32		nref_max;
	s32		nref_min;
	s32		nvco;
	s32		nvco_max;
	s32		nvco_min;
	s32		pll;
} fsamp_t;



//*****************************************************************************
static void _dsi_fsamp_compute_legacy(fsamp_t* data)
{
	double	delta;
	double	fgen;
	s32		fgen_max	= data->fgen_max;
	s32		fgen_min	= data->fgen_min;
	double	fsamp_got;
	s32		fsamp_max	= data->fsamp_max;
	s32		fsamp_min	= data->fsamp_min;
	double	fsamp_want	= (double) data->fsamp_want;
	int		init		= 0;	// Have we performed the first run initialization?
	s32		ndiv;
	s32		ndiv_max	= data->ndiv_max;
	s32		ndiv_min	= data->ndiv_min;
	s32		nrate;
	s32		nrate_max	= data->nrate_max;
	s32		nrate_min	= data->nrate_min;

	for (nrate = nrate_min; nrate <= nrate_max; nrate++)
	{
		fgen	= DSI_FGEN_LEGACY_HZ(nrate);

		if ((init) && ((fgen < fgen_min) || (fgen > fgen_max)))
			continue;

		for (ndiv = ndiv_min; ndiv <= ndiv_max; ndiv++)
		{
			fsamp_got	= DSI_FSAMP_LEGACY_SPS(nrate, ndiv);

			if ((init) && ((fsamp_got < fsamp_min) || (fsamp_got > fsamp_max)))
				continue;

			delta	= fsamp_got - fsamp_want;

			if ((init == 0) || ((ABS(delta)) < (ABS(data->delta))))
			{
				init			= 1;
				data->nrate		= nrate;
				data->fgen		= fgen;
				data->ndiv		= ndiv;
				data->fsamp_got	= fsamp_got;
				data->delta		= delta;
			}
		}
	}
}



//*****************************************************************************
static void _dsi_fsamp_compute_pll(fsamp_t* data)
{
	double	delta;
	double	fgen;
	s32		fgen_max	= data->fgen_max;
	s32		fgen_min	= data->fgen_min;
	s32		fref		= data->fref;
	double	fsamp_got;
	s32		fsamp_max	= data->fsamp_max;
	s32		fsamp_min	= data->fsamp_min;
	double	fsamp_want	= (double) data->fsamp_want;
	int		init		= 0;	// Have we performed the first run initialization?
	s32		ndiv;
	s32		ndiv_max	= data->ndiv_max;
	s32		ndiv_min	= data->ndiv_min;
	s32		nref;
	s32		nref_max	= data->nref_max;
	s32		nref_min	= data->nref_min;
	s32		nvco;
	s32		nvco_max	= data->nvco_max;
	s32		nvco_min	= data->nvco_min;

	for (nvco = nvco_min; nvco <= nvco_max; nvco++)
	{
		for (nref = nref_min; nref <= nref_max; nref++)
		{
			fgen	= DSI_FGEN_PLL_HZ(fref, nvco, nref);

			if ((init) && ((fgen < fgen_min) || (fgen > fgen_max)))
				continue;

			for (ndiv = ndiv_min; ndiv <= ndiv_max; ndiv++)
			{
				fsamp_got	= DSI_FSAMP_PLL_SPS(fref, nvco, nref, ndiv);

				if ((init) && ((fsamp_got < fsamp_min) || (fsamp_got > fsamp_max)))
					continue;

				delta	= fsamp_got - fsamp_want;

				if ((init == 0) || ((ABS(delta)) < (ABS(data->delta))))
				{
					init			= 1;
					data->nvco		= nvco;
					data->nref		= nref;
					data->fgen		= fgen;
					data->ndiv		= ndiv;
					data->fsamp_got	= fsamp_got;
					data->delta		= delta;
				}
			}
		}
	}
}



/******************************************************************************
*
*	Function:	dsi_fsamp_ai_compute
*
*	Purpose:
*
*		Calculate the best values for the specified sample rate.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely? If not, then work quietly.
*
*		options	Display all computed options within this range of the requested
*				sample rate.
*
*		fref	The is the Fref value. This is ignored if this is a legacy
*				device.
*
*		fsamp	The desired sample rate.
*
*		nrate	Put the resulting Nrate value here.
*
*		ndiv	Put the resulting Ndiv value here.
*
*		rate	Report the sample rate achieved here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

int dsi_fsamp_ai_compute(
	int		fd,
	int		index,
	int		verbose,
	s32		fref,
	s32		fsamp,
	s32*	nvco,
	s32*	nref,
	s32*	nrate,
	s32*	ndiv,
	double*	rate)
{
	fsamp_t	data;
	int		errs	= 0;

	if (verbose)
	{
		gsc_label_index("AI Sample Rate", index);
		gsc_label_long_comma(fsamp);
		printf(" S/S");
	}

	memset(&data, 0, sizeof(data));
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_FGEN_MAX, &data.fgen_max);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_FGEN_MIN, &data.fgen_min);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_FSAMP_MAX, &data.fsamp_max);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_FSAMP_MIN, &data.fsamp_min);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_NDIV_MAX, &data.ndiv_max);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_NDIV_MIN, &data.ndiv_min);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_NRATE_MAX, &data.nrate_max);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_NRATE_MIN, &data.nrate_min);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_NVCO_MAX, &data.nvco_max);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_NVCO_MIN, &data.nvco_min);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_PLL_PRESENT, &data.pll);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_NREF_MAX, &data.nref_max);
	errs	+= dsi_query(fd, -1, 0, DSI_QUERY_NREF_MIN, &data.nref_min);

	data.delta		= FLT_MAX;
	data.fref		= fref;
	data.fsamp_want	= LIMIT_RANGE(fsamp, data.fsamp_min, data.fsamp_max);

	if (data.pll)
	{
		_dsi_fsamp_compute_pll(&data);

		if (verbose)
		{
			printf("%s  (PLL: ", errs ? " FAIL <---" : "");
			printf("Fref ");
			gsc_label_long_comma((long) fref);
			printf(", Nvco %ld", (long) data.nvco);
			printf(", Nref %ld", (long) data.nref);
			printf(", Ndiv %ld", (long) data.ndiv);
			printf(", Fsamp %.3lf", (double) data.fsamp_got);
			printf(")\n");
		}
	}
	else
	{
		_dsi_fsamp_compute_legacy(&data);

		if (verbose)
		{
			printf("%s  (Legacy: ", errs ? " FAIL <---" : "");
			printf("Fref ");
			gsc_label_long_comma((long) fref);
			printf(", Nrate %ld", (long) data.nrate);
			printf(", Ndiv %ld", (long) data.ndiv);
			printf(", Fsamp %.3lf", (double) data.fsamp_got);
			printf(")\n");
		}
	}

	if (nvco)
		nvco[0]		= data.nvco;

	if (nref)
		nref[0]		= data.nref;

	if (nrate)
		nrate[0]	= data.nrate;

	if (ndiv)
		ndiv[0]		= data.ndiv;

	if (rate)
		rate[0]		= data.fsamp_got;

	return(errs);
}



/******************************************************************************
*
*	Function:	dsi_fsamp_ai_report
*
*	Purpose:
*
*		Determine and report the sample rate for the specified channel.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely? If not, then work quietly.
*
*		chan	The index of the channel of interest.
*
*		fref	The user specified reference frequency.
*
*		sps		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int dsi_fsamp_ai_report(int fd, int index, int verbose, int chan, s32* fref, s32* sps)
{
	char	buf[64];
	int		c_p_g;		// channels per group
	s32		chans;
	s32		dividers	= 0;
	int		errs		= 0;
	s32		fsamp		= 0;
	int		group;
	s32		groups;
	s32		grp0rar;			// Ch Group 0 has precidence in RAR
	s32		ndiv		= 0;
	s32		nrate		= 0;
	s32		nref		= 0;
	s32		nvco		= 0;
	s32		order;
	s32		pll			= 0;
	s32		src			= -1;
	s32		tmp			= -1;	// in case fref == NULL

	if (verbose)
	{
		sprintf(buf, "Channel %d", chan);
		gsc_label_index(buf, index);
	}

	if (fref == NULL)
		fref	= &tmp;

	for (;;)
	{
		if ((chan < 0) || (chan > 32))
		{
			errs	= 1;

			if (verbose)
				printf("FAIL <---  (invalid channel index: %d)\n", chan);

			break;
		}

		errs	+= dsi_channel_order(fd, -1, 0, -1, &order);
		errs	+= dsi_query(fd, -1, 0, DSI_QUERY_CH_GRP_0_RAR, &grp0rar);
		errs	+= dsi_query(fd, -1, 0, DSI_QUERY_PLL_PRESENT, &pll);
		errs	+= dsi_query(fd, -1, 0, DSI_QUERY_CHANNEL_QTY, &chans);
		errs	+= dsi_query(fd, -1, 0, DSI_QUERY_CHANNEL_GPS, &groups);
		errs	+= dsi_query(fd, -1, 0, DSI_QUERY_RATE_DIV_QTY, &dividers);
		c_p_g	= chans / groups;
		group	= chan / c_p_g;
		errs	+= dsi_channel_group_source(fd, -1, 0, group, -1, &src);

		if ((src == DSI_CH_GRP_SRC_ENABLE)	||
			(order == DSI_CHANNEL_ORDER_SYNC))
		{
			// When we get an ENABLE source we must use the channel 0 source.
			// For synchronous ordering we also use the channel 0 source.
			errs	+= dsi_channel_group_source(fd, -1, 0, 0, -1, &src);
		}

		switch (src)
		{
			default:
			case DSI_CH_GRP_SRC_ENABLE:

				errs	= 1;

				if (verbose)
				{
					printf(	"FAIL <---  (invalid logic or source option: %ld)\n",
							(long) src);
				}

				break;

			case DSI_CH_GRP_SRC_GEN_A:

				if (pll)
				{
					errs	+= dsi_rate_gen_x_nvco(fd, -1, 0, 0, -1, &nvco);
					errs	+= dsi_rate_gen_x_nref(fd, -1, 0, 0, -1, &nref);
					errs	+= dsi_fref_compute(fd, -1, 0, fref);
				}
				else
				{
					errs	+= dsi_rate_gen_x_nrate(fd, -1, 0, 0, -1, &nrate);
				}

				errs	+= dsi_rate_div_x_ndiv(fd, -1, 0, 0, -1, &ndiv);
				break;

			case DSI_CH_GRP_SRC_GEN_B:

				if (pll)
				{
					errs	+= dsi_rate_gen_x_nvco(fd, -1, 0, 1, -1, &nvco);
					errs	+= dsi_rate_gen_x_nref(fd, -1, 0, 1, -1, &nref);
					errs	+= dsi_fref_compute(fd, -1, 0, fref);
				}
				else
				{
					errs	+= dsi_rate_gen_x_nrate(fd, -1, 0, 1, -1, &nrate);
				}

				errs	+= dsi_rate_div_x_ndiv(fd, -1, 0, 0, -1, &ndiv);
				break;

			case DSI_CH_GRP_SRC_EXTERN:
			case DSI_CH_GRP_SRC_DIR_EXTERN:
			case DSI_CH_GRP_SRC_DISABLE:

				break;
		}

		if (errs)
		{
		}
		else if (src == DSI_CH_GRP_SRC_DISABLE)
		{
			fsamp	= 0;

			if (verbose)
				printf("Disabled\n");
		}
		else if (	(src == DSI_CH_GRP_SRC_EXTERN)	||
					(src == DSI_CH_GRP_SRC_DIR_EXTERN))
		{
			fsamp	= -1;

			if (verbose)
				printf("External Source\n");
		}
		else if (pll)
		{
			fsamp	= DSI_FSAMP_PLL_SPS(fref[0], nvco, nref, ndiv);

			if (verbose)
			{
				gsc_label_long_comma(fsamp);
				printf(" S/S  (Fref ");
				gsc_label_long_comma(fref[0]);
				printf(" Hz, Nvco ");
				gsc_label_long_comma(nvco);
				printf(", Nref ");
				gsc_label_long_comma(nref);
				printf(", Ndiv ");
				gsc_label_long_comma(ndiv);
				printf(")\n");
			}
		}
		else
		{
			fsamp	= DSI_FSAMP_LEGACY_SPS(nrate, ndiv);

			if (verbose)
			{
				gsc_label_long_comma(fsamp);
				printf(" S/S  (Fref ");
				gsc_label_long_comma(fref[0]);
				printf(" Hz, Nrate ");
				gsc_label_long_comma(nrate);
				printf(", Ndiv ");
				gsc_label_long_comma(ndiv);
				printf(")\n");
			}
		}

		break;
	}

	if (sps)
		sps[0]	= fsamp;

	return(errs);
}



/******************************************************************************
*
*	Function:	dsi_fsamp_ai_report_all
*
*	Purpose:
*
*		Determine and report the sample rate for all channels.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely? If not, then work quietly.
*
*		fref	The Fref is reported here.
*
*		fsamp	The overall sampe rate is reported here.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int dsi_fsamp_ai_report_all(int fd, int index, int verbose, s32* fref, s32* fsamp)
{
	s32	chans;
	int	chan;
	int	errs;
	s32	sps;
	s32	total	= 0;

	if (verbose)
		gsc_label_index("Sample Rates", index);

	errs	= dsi_query(fd, index, 0, DSI_QUERY_CHANNEL_QTY, &chans);

	if (errs == 0)
	{
		if (verbose)
			printf("\n");

		gsc_label_level_inc();

		for (chan = 0; chan < chans; chan++)
		{
			errs	+= dsi_fsamp_ai_report(fd, index, verbose, chan, fref, &sps);
			total	+= sps;
		}

		if (verbose)
		{
			gsc_label("Overall Rate");
			gsc_label_long_comma(total);
			printf(" S/S\n");
		}

		gsc_label_level_dec();

		if (fsamp)
			fsamp[0]	= total;
	}

	return(errs);
}


